/*****************************************************************************

  The following code is derived, directly or indirectly, from the SystemC
  source code Copyright (c) 1996-2006 by all Contributors.
  All Rights reserved.

  The contents of this file are subject to the restrictions and limitations
  set forth in the SystemC Open Source License Version 2.4 (the "License");
  You may not use this file except in compliance with such restrictions and
  limitations. You may thread_hobtain instructions on how to receive a copy of the
  License at http://www.systemc.org/. Software distributed by Contributors
  under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
  ANY KIND, either express or implied. See the License for the specific
  language governing rights and limitations under the License.

 *****************************************************************************/

/*****************************************************************************
 
  bios.cpp -- System Bios.
 
  Original Author: Martin Wang, Synopsys, Inc.
 
 *****************************************************************************/
 
/*****************************************************************************
 
  MODIFICATION LOG - modifiers, enter your name, affiliation, date and
  changes you are making here.
 
      Name, Affiliation, Date:
  Description of Modification:
 
 *****************************************************************************/

#include "src/systemc.h"
#include "bios.h"
#include "directive.h"
#include "work_load.h"
#include "SystemC/os_systemc.h"

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>



#define bios_memory_size 40000
// These variables are used for priniting only.
// Printing is performed when the BIOS unit is not active.
extern "C" bool	 mainPrintBiosFlag ; 


unsigned bios_address ;
unsigned bios_instruction ; 
unsigned imemory[bios_memory_size];
unsigned itagmemory[bios_memory_size];

void biosPrint ()
{  
	if ( mainPrintBiosFlag == true ) 
	{
		mainPrintBiosFlag = false ; 
		if (PRINT_BIOS)
		{

			System_printf( "BIOS: Fetching Memory(%0d):" , bios_address ) ;
			System_printf( " (0%h)", bios_instruction, "\n")  ;

		}
	}
}

void bios_init_variables ()
{
		// initialize instruction imemory from external file
	FILE *fp = fopen("bios","r");
	int size=0;
	int mem_word;

	System_printf("** ALERT ** BIOS: initialize BIOS\n");
	for (size = 0; size < bios_memory_size; size++) {	// initialize bad data
		imemory[size] = 0xffffffff;
		itagmemory[size] = 0xffffffff;
	}
	size = 0;
	while (fscanf(fp,"%x\n", &mem_word) != EOF) 
	{
		imemory[size] = mem_word;
		itagmemory[size] = size;
		size++;
	}
}
void bios::entry_bios()
{

bios_init_variables (); 
work_load ( INFINITE, ThreadBIOSID) ;
sc_module::wait() ;  
while (true) 
{   
	while ( cs.read() == false ) // Chip select of the RAM.

	{
	   
		work_load ( INFINITE, ThreadBIOSID) ; 
		sc_module::wait() ; // os_systemcWait( INFINITE ) ;
	}		
		
		
    	bios_address = addr.read();
		if (bios_address< BOOT_LENGTH) 
		{		// in BOOTING STAGE
    		if (we.read() == true) 
			{ // Write operation
				
      			work_load ( INFINITE, ThreadBIOSID) ; 
				sc_module::wait(bios_wait_cycles-1 ) ;
				bios_instruction = datain.read();
      			imemory[bios_address] = bios_instruction ; 
    		}
    		else 
			{     
				// Read operation
      			if (bios_wait_cycles > 2)
      			{
					work_load ( INFINITE, ThreadBIOSID) ; 
					sc_module::wait(bios_wait_cycles-2) ;  // Introduce delay needed
      			}
      			dataout.write(imemory[bios_address]);
				    mainPrintBiosFlag  = true ; 
			
					bios_valid.write(true);
					work_load ( INFINITE, ThreadBIOSID) ;
      				sc_module:: wait() ; //  os_systemcWait( INFINITE ) ;
					bios_valid.write(false);
					work_load ( INFINITE, ThreadBIOSID) ;
					sc_module::wait() ; //  os_systemcWait( INFINITE ) ;
	    		}    
		} 
		else 
		{
				bios_valid.write(false);
			work_load ( INFINITE, ThreadBIOSID) ;
				sc_module::wait() ; // os_systemcWait( INFINITE ) ;

		}
		
	}
} // End of BIOS entry function

