


/*****************************************************************************

  The following code is derived, directly or indirectly, from the SystemC
  source code Copyright (c) 1996-2006 by all Contributors.
  All Rights reserved.

  The contents of this file are subject to the restrictions and limitations
  set forth in the SystemC Open Source License Version 2.4 (the "License");
  You may not use this file except in compliance with such restrictions and
  limitations. You may obtain instructions on how to receive a copy of the
  License at htthread_htp://www.systemc.org/. Software distributed by Contributors
  under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF = 0 )
  ANY KIND, either express or implied. See the License for the specific
  language governing rights and limitations under the License.

 *****************************************************************************/

/*****************************************************************************
 
  decode.cpp  -- Instruction Decode Unit.
 
  Original Author: Martin Wang, Synopsys, Inc.
 
 *****************************************************************************/
 
/*****************************************************************************
 
  MODIFICATION LOG - modifiers, enter your name, affiliation, date and
  changes you are making here.
 
      Name, Affiliation, Date:
  Description of Modification:
 
 *****************************************************************************/
 

//#include <climits>	// for definition on value 's MAX
#include "src/systemc.h"
#include "decode.h"
#include "threads.h"
#include "work_load.h"
#include "SystemC/os_systemc.h"
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>
extern "C" bool mainPrintDecodeFlag ;
extern "C" bool mainDecodePintOutValidFlag ;
extern "C" bool mainPrintDecodeDumpRegistersFlag  ;
extern "C" bool mainPrintPerformanceFlag ; 
extern "C" bool mainPrintDecodeShutDownFlag ; 

signed int cpu_reg[32];			//CPU register
signed int vcpu_reg[32];		//virtual CPU register
bool 	   cpu_reg_lock[32];		//lock architectural state register
unsigned int pc_reg;			//pc register
unsigned int jalpc_reg;			//jump back register

void decode_init() 
{


	    FILE *fp = fopen("register","r");
        int size=0;
        int mem_word;
        System_printf("** ALERT ** Instruction Decode: initialize Architectural Registers\n");
        while (fscanf(fp,"%x", &mem_word) != EOF) {
                cpu_reg[size] = mem_word;
		size++;
	}
	pc_reg = 0;
	jalpc_reg = 0;
	for (int j =0; j<32; j++) vcpu_reg[j] 	= 0;
	for (int k =0; k<32; k++) cpu_reg_lock[k] = 0;
}


// extern  signed int cpu_reg[32];	 
void mainPrintPerformance () ;

 unsigned int 	instr_tmp = 0;
  unsigned int 	opcode_tmp = 0;
  unsigned int  regA_tmp = 0;
  unsigned int  regB_tmp = 0;
  unsigned int  regC_tmp = 0;
  unsigned int  imm_tmp = 0;
  unsigned int  offset_tmp = 0;
  signed   int  label_tmp = 0;
  unsigned int  longlabel_tmp = 0;
  unsigned int  lastreg_tmp = 0;
  signed int	srcA_tmp = 0;
  signed int	srcB_tmp = 0;
  signed int	srcC_tmp = 0;
  int		i;
  bool		branch_direction_tmp = 0;

void DecodePrintDumpedRegisters()
{ int i ; 
	if (mainPrintDecodeDumpRegistersFlag == true )
	{
		mainPrintDecodeDumpRegistersFlag = false ; 
	
				  
		System_printf("\n\n\t*******************************\n");
		System_printf("\tInstruction Decode Unit: Registers Dump\n");
		System_printf("REG :==================================================================\n");
		for(i =0; i<32; i++)
		{
			System_printf("  R%2d(%08x)  ",i, cpu_reg[i]);
	
			if (i%4==3)
			{
						System_printf("\n");
			}
		}
		System_printf("=======================================================================\n\n");
	}
}

void decode_print_variables ()
{
		
  if (mainPrintDecodeFlag == true ) 
  {
	  mainPrintDecodeFlag = false ; 
	System_printf("IDU opcode = %0x regC = %0x regA = %0x regB = %0x\n",opcode_tmp, regC_tmp, regA_tmp, 
			 regB_tmp, "\n" );

		
       // lastreg_tmp = regC_tmp;
		System_printf( "\t------------------------------- \n" ) ;

		switch(opcode_tmp) 
		{
			case 0x0:	// halt
			
			//	mainPrintDecodeDumpRegisters () ;
				break;
			case 0x01:	// add R1, R2, R3
				System_printf("\tID: R%0x= R%0x(=%0x)+R%0x(=%0x)", 
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
					
				break;
			case 0x02:	// addi R1, R2, #value
				System_printf("\tID: R%0x= R%0x(=%0x)+%0x", regC_tmp, regA_tmp, srcA_tmp, imm_tmp);
				break;
			case 0x03:	// addc R1, R2, R3 + Carrybit
				System_printf("\tID: R%0x=  R%0x(=%0x)+R%0x(=%0x)+C", 
						regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				break;
			case 0x04:	// sub R1, R2, R3 
				System_printf("\tID: R%0x=R%0x(=%0x)-R%0x(=%0x)", 
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				break;
			case 0x05:	// subi R1, R2,  #value
				System_printf("\tID: R%0x=R%0x(=%0x)-%0x", regC_tmp, regA_tmp, srcA_tmp, imm_tmp);
				break;
			case 0x06:	// subc R1, R2, R3 - Carrybit
                System_printf("\tID: R%0x=R%0x(=%0x)-R%0x(=%0x)-C", 
						regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				break;
			case 0x07:	// mul R1, R2, R3 
				System_printf("\tID: R%0x=R%0x(=%0x)*R%0x(=%0x)", 	
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				
				break;
			case 0x08:	// div R1, R2, R3 
				System_printf("\tID: R%0x=R%0x(=%0x)/R%0x(=%0x)", 	
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				break;
			case 0x09:	// nand R1, R2, R3 
				System_printf("\tID: R%0x=R%0x(=%x) nand R%0x(=%x)", 	
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				break;
			case 0x0a:	// and R1, R2, R3 
				System_printf("\tID: R%0x=R%0x(=%x) and R%0x(=%x)", 	
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				break;
			case 0x0b:	// or R1, R2, R3 
                   System_printf("\tID: R%0x=R%0x(=%x) or R%0x(=%x)",
                          regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
				   break;
			case 0x0c:	// xor R1, R2, R3 
                System_printf("\tID: R%0x=R%0x(=%x) xor R%0x(=%x)",
                   regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
			    break;
			case 0x0d:      // NOT R1, R2 
                System_printf("\tID: R%0x= NOT R%0x(=%x)", regC_tmp, regA_tmp, srcA_tmp);   
               break;
			case 0x0e:      // modulo R1 = R2 mod R3 
               System_printf("\tID: R%0x= R%0x(=%x) mod R%0x(=%x)", 
					regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
               break;
			case 0x4d:	// lw R1, R2, offset
				System_printf("\tID: R%0x<=mem[R%0x=(%0x)+%0x]", regC_tmp, regA_tmp, srcA_tmp, offset_tmp);
				break;
			case 0x4e:	// sw R1, R2, offset
				System_printf("\tID: R%0x>=mem[R%0x(=%0x) + %0x]", regC_tmp, regA_tmp, srcA_tmp,offset_tmp);
				break;
			case 0x0f:	// mov R1, R2
				System_printf("\tID: R%0x=R%0x(=%0x)", regC_tmp, regA_tmp, srcA_tmp);
				break;
			case 0x10:	// beq R1, R2, label
				if (srcC_tmp == srcA_tmp) 
				{
					System_printf("\tID: beq R%0x(=%0x), R%0x(=%0x), pc+=(%0x).\n", 
						regC_tmp, srcC_tmp, regA_tmp, srcA_tmp, label_tmp);
				} else
				{
					System_printf("\tID: beq R%0x(=%0x) != R%0x(=%0x),pc++.\n", 
						regC_tmp, srcC_tmp, regA_tmp, srcA_tmp);
				}
				break;
			case 0x11:	// bne R1, R2, label
               if (srcC_tmp != srcA_tmp) 
			   {
			      System_printf("\tID: bne R%0x(=%0x), R%0x(=%0x), pc+=(%0x).\n",
                  regC_tmp, srcC_tmp, regA_tmp, srcA_tmp, label_tmp);
               } 
			   else
			   {
                   System_printf("\tID: bne R%0x(=%0x) = R%0x(=%0x),pc++.\n",
                         regC_tmp, srcC_tmp, regA_tmp, srcA_tmp);
               }
               break;
			case 0x12:	// bgt R1, R2, label
                  if (srcC_tmp > srcA_tmp) 
				  {
	                 System_printf("\tID: bgt R%0x(=%0x)>R%0x(=%0x), pc+=(%0x).\n",
                         regC_tmp, srcC_tmp, regA_tmp, srcA_tmp, label_tmp);
                  } 
				  else 
				  {
                        System_printf("\tID: bgt R%0x(=%0x) <= R%0x(=%0x),pc++.\n",
                                                    regC_tmp, srcC_tmp, regA_tmp, srcA_tmp);
                  }
			      break;
			case 0x13:	// bge R1, R2, label
                   if (srcC_tmp >= srcA_tmp)
				   {
					  System_printf("\tID: bge R%0x(=%0x)>=R%0x(=%0x), pc+=(%0x).\n",
                                 regC_tmp, srcC_tmp, regA_tmp, srcA_tmp, label_tmp);
				   }
				   else 
				   {
                      System_printf("\tID: bge R%0x(=%0x) < R%0x(=%0x),pc++.\n",
                                                        regC_tmp, srcC_tmp, regA_tmp, srcA_tmp);
                    }
			        break;
			case 0x14:	// blt R1, R2, label
          
                 if (srcC_tmp < srcA_tmp)
				 {
					 System_printf("\tID: blt R%0x(=%0x)<R%0x(=%0x), pc+=(%0x).\n",
                               regC_tmp, srcC_tmp, regA_tmp, srcA_tmp, label_tmp);
                 } 
				 else 
				 {
                       System_printf("\tID: blt R%0x(=%0x) >= R%0x(=%0x), pc++.\n",
                         regC_tmp, srcC_tmp, regA_tmp, srcA_tmp);
                 }
                 break;
			case 0x15:	// ble R1, R2, label
                 if (srcC_tmp <= srcA_tmp)
				 {
					
                    System_printf("\tID: ble R%0x(=%0x)<=R%0x(=%0x), pc+=(%0x).\n",
                                      regC_tmp, srcC_tmp, regA_tmp, srcA_tmp, label_tmp);
                  } 
				 else 
				 {
                            System_printf("\tID: ble R%0x(=%0x)>R%0x(=%0x), pc++.\n",
                                       regC_tmp, srcC_tmp, regA_tmp, srcA_tmp);
                 }
									
				 break;
			case 0x16:	// j label
                 System_printf("\tID: pc jump to >= (%0x).", longlabel_tmp );
						 
                  break;
			case 0x17:	// jal label for procedure call
					System_printf("\tID: j pc(%0x) and sp <- pc", longlabel_tmp );
				
			case 0x18:	// jr
					System_printf("\tID: jr and pc <- sp\n");
					break;
            case 0x29:      // fadd R1, R2, R3
                   System_printf("\tID: FR%0x=FR%0x(=%x)+FR%0x(=%x)",
                          regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
									
                    break;
                        
			case 0x2a:      // fsub R1, R2, R3
                  System_printf("\tID: FR%0x=FR%0x(=%x)-FR%0x(=%x)",
                                                regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
										
                  break;
             case 0x2b:      // fmul R1, R2, R3
                  System_printf("\tID: FR%0x=FR%0x(=%x)*FR%0x(=%x)",
                             regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
							
                   break;
             case 0x2c:      // fdiv R1, R2, R3
                  System_printf("\tID: FR%0x=FR%0x(=%x)/FR%0x(=%x)",
                                                regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
								
                 break;
             case 0x31:      // mmxadd R1, R2, R3
                 System_printf("\tID: MR%0x=MR%0x(=%x)+MR%0x(=%x)",
                                              regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
									
                 break;
             case 0x32:      // mmxadds R1, R2, R3
                  System_printf("\tID: MR%0x=MR%0x(=%x)+MR%0x(=%x) Sat",
                                                regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
						 break;
                   case 0x33:      // mmxsub R1, R2, R3
                         System_printf("\tID: MR%0x=MR%0x(=%x)-MR%0x(=%x)",
                                             regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
						break;
                    case 0x34:      // mmxsubs R1, R2, R3
                           System_printf("\tID: MR%0x=MR%0x(=%x)-MR%0x(=%x) Sat",
                                     regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
						 break;
                    case 0x35:      // pmadd R1, R2, R3
                                        System_printf("\tID: MR%0x=MR%0x(=%x) ,MR%0x(=%x) PMADD",
                                                regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
									
                                break;
                    case 0x36:      // pack R1, R2, R3
                                        System_printf("\tID: MR%0x=MR%0x(=%x) pack MR%0x(=%x)",
                                                regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
								
                                        break;
                    case 0x37:      // mmxck R1, R2, R3 MMX Chroma Keying
                                  System_printf("\tID: MR%0x=MR%0x(=%x) mmxck MR%0x(=%x)",
                                              regC_tmp, regA_tmp, srcA_tmp, regB_tmp, srcB_tmp);
									
                                        break;
                        case 0xe0:      // flush register
                                        System_printf("\tID: flush all registers");
							
                                        break;
                        case 0xf0:      // ldpid process_number
                                        System_printf("\tID: ld pid =%0x", longlabel_tmp );
								
                                        break;
			case 0xf1:	// movi R1, #value
				System_printf("\tID: R%0x=%0x", 	regC_tmp, imm_tmp);
					
					break;
                        case 0xff:      // QUIT
					mainPrintDecodeShutDownFlag = true ; 
					mainPrintDecodeDumpRegistersFlag = true ; 
					mainPrintPerformanceFlag = true ; 
					DecodePrintDumpedRegisters () ; 
					mainPrintPerformance () ;
					sc_stop();
					
                   
					break;
			default :
					System_printf("\tInstruction Decoder: INVALID OPCODE");
					mainPrintDecodeShutDownFlag = true ; 
					mainPrintDecodeDumpRegistersFlag = true ; 
					mainPrintPerformanceFlag = true ; 
					DecodePrintDumpedRegisters () ; 
					mainPrintPerformance () ;
					sc_stop();
				
					break;
		}
	}
}




void decode::entry_decode()
{ 
	// ThreadDecode () ;
	decode_init () ; 
	work_load (INFINITE, ThreadDecodeID)  ;
	sc_module::wait() ; 

 branch_valid.write(false);

 decode_valid.write(false);

 float_valid.write(false);

 mmx_valid.write(false);
  work_load (INFINITE, ThreadDecodeID)  ; 
 sc_module::wait( 2 );

  while (true)
  {
	if (destreg_write.read() == true) 
	{
		cpu_reg[destreg_write_src.read()] = alu_dataout.read();
    
	}
	if (dram_rd_valid.read() == true) 
	{
		cpu_reg[dram_write_src.read()] = dram_dataout.read();
		mainDecodePintOutValidFlag = true ; 


	}
	if (fpu_valid.read() == true) 
	{
		cpu_reg[fpu_destout.read()] = fpu_dout.read();
  
	}
	if (instruction_valid.read() == true)
	{
		pc_reg = pc.read();
		if (clear_branch.read() == true) {

			branch_valid.write(false);
		}
   		instr_tmp  = instruction.read();
		opcode_tmp 	= (instr_tmp & 0xff000000) >> 24; 
		regC_tmp   	= (instr_tmp & 0x00f00000) >> 20; 
		regA_tmp   	= (instr_tmp & 0x000f0000) >> 16; 
		regB_tmp   	= (instr_tmp & 0x0000f000) >> 12; 
		imm_tmp    	= (instr_tmp & 0x0000ffff); 
		offset_tmp 	= (instr_tmp & 0x00000fff); 
		label_tmp  	= (instr_tmp & 0x0000ffff); 
		longlabel_tmp   = (instr_tmp & 0x00ffffff); 
		branch_direction_tmp = (instr_tmp & 0x00008000) >> 15;
		if (branch_direction_tmp) {			// handle backward branch
			label_tmp = - (0xffff - label_tmp + 1)  ;
		}
		
		srcA_tmp = cpu_reg[regA_tmp];
		srcB_tmp = cpu_reg[regB_tmp];
		srcC_tmp = cpu_reg[regC_tmp];
		  work_load (INFINITE, ThreadDecodeID)  ; 
		sc_module::wait() ;  ;
                lastreg_tmp = regC_tmp;
        /******************************************************************************
		if (regA_tmp == lastreg_tmp){
			forward_A.write(true);
			forward_B.write(false);
		}  else  if (regB_tmp == lastreg_tmp){
				forward_A.write(false);
				forward_B.write(true);
		} else {
			forward_A.write(false);
			forward_B.write(false);
		}
*********************************************************************************/
		mainPrintDecodeFlag = true ; 
		switch(opcode_tmp) {
			case 0x0:	// halt
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					break;
			case 0x01:	// add R1, R2, R3
					src_A.write(srcA_tmp);
					src_B.write(srcB_tmp);
					alu_src.write(regC_tmp);	
					alu_op.write(3);	
					decode_valid.write(true);
					   work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					decode_valid.write(false);
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					break;
			case 0x02:	// addi R1, R2, #value
					src_A.write(srcA_tmp);
					src_B.write(imm_tmp);
					alu_src.write(regC_tmp);
					alu_op.write(3);	
					decode_valid.write(true);
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					decode_valid.write(false);
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					break;
			case 0x03:	// addc R1, R2, R3 + Carrybit
					src_A.write(srcA_tmp);
					src_B.write(srcB_tmp);
					alu_src.write(regC_tmp);
					alu_op.write(1);
					decode_valid.write(true);
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					decode_valid.write(false);
					  work_load (INFINITE, ThreadDecodeID)  ; 
					sc_module::wait() ;  ;
					break;
			case 0x04:	// sub R1, R2, R3 
				   src_A.write(srcA_tmp);
                   src_B.write(srcB_tmp);
                   alu_src.write(regC_tmp);
                   alu_op.write(4);
                   decode_valid.write(true);
				   work_load (INFINITE, ThreadDecodeID)  ; 
                  work_load (INFINITE, ThreadDecodeID)  ;    sc_module::wait();
                   decode_valid.write(false);
				  work_load (INFINITE, ThreadDecodeID)  ;            
				  work_load (INFINITE, ThreadDecodeID)  ;    sc_module::wait();
					break;
			case 0x05:	// subi R1, R2,  #value
			     src_A.write(srcA_tmp);
                 src_B.write(imm_tmp);
                 alu_src.write(regC_tmp);
                 alu_op.write(4);
                 decode_valid.write(true);
                work_load (INFINITE, ThreadDecodeID)  ;     sc_module::wait();
                 decode_valid.write(false);
                work_load (INFINITE, ThreadDecodeID)  ;     sc_module::wait();
				break;
			case 0x06:	// subc R1, R2, R3 - Carrybit
                src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(2);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;     sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;     sc_module::wait();
				break;
			case 0x07:	// mul R1, R2, R3 
                src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(5);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;     sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;   ;    sc_module::wait();
				break;
			case 0x08:	// div R1, R2, R3 
                src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(6);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;   ;    sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;   ;    sc_module::wait();
				break;
			case 0x09:	// nand R1, R2, R3 
				src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(7);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
				break;
			case 0x0a:	// and R1, R2, R3 
	            src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(8);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
				break;
			case 0x0b:	// or R1, R2, R3 
                src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(9);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                break;
			case 0x0c:	// xor R1, R2, R3 

                src_A.write(srcA_tmp);
                src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(10);
                decode_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                decode_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                break;
            case 0x0d:      // NOT R1, R2 
                                       
                 src_A.write(srcA_tmp);
                 src_B.write(0);
                 alu_src.write(regC_tmp);
                 alu_op.write(11);
                 decode_valid.write(true);
                work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                 decode_valid.write(false);
                work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                 break;
            case 0x0e:      // modulo R1 = R2 mod R3 
                 src_A.write(srcA_tmp);
                 src_B.write(srcB_tmp);
                 alu_src.write(regC_tmp);
                 alu_op.write(14);
                 decode_valid.write(true);
                work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                 decode_valid.write(false);
                work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                 break;
		   case 0x4d:	// lw R1, R2, offset
			    mem_access.write(true);
				mem_write.write(false);
				alu_src.write(regC_tmp);
				offset_tmp = offset_tmp + srcA_tmp;
				mem_address.write(offset_tmp);
				sc_module::wait(); ;
				mem_access.write(false);
				sc_module::wait(); ;
				break;
			case 0x4e:	// sw R1, R2, offset
				mem_access.write(true);
				mem_write.write(true);
				offset_tmp = offset_tmp + srcA_tmp;
				mem_address.write(offset_tmp);
				sc_module::wait(); ;
				mem_access.write(false);
				mem_write.write(false);
				sc_module::wait(); ;
				break;
			case 0x0f:	// mov R1, R2
				src_A.write(srcA_tmp);
				src_B.write(0);
				alu_src.write(regC_tmp);
				alu_op.write(3);	
				decode_valid.write(true);
				sc_module::wait(); ;
				decode_valid.write(false);
				sc_module::wait(); ;
				break;
			case 0x10:	// beq R1, R2, label
					src_A.write(0);
					src_B.write(0);
					alu_src.write(0);
					alu_op.write(3);	
					decode_valid.write(true);
					if (srcC_tmp == srcA_tmp) 
					{
						branch_target_address.write(pc_reg + label_tmp);
						br_instruction_address.write(instr_tmp);
						branch_valid.write(true);
					
					}
					else
					{

					}
					sc_module::wait(); ;
					branch_target_address.write(pc_reg + 1);
					decode_valid.write(false);
					sc_module::wait(); ;
					break;
			case 0x11:	// bne R1, R2, label
                   src_A.write(0);
                   src_B.write(0);
                   alu_src.write(0);
                   alu_op.write(3);
                   decode_valid.write(true);
                   if (srcC_tmp != srcA_tmp)
				   {
						branch_target_address.write(pc_reg + label_tmp);
						br_instruction_address.write(instr_tmp);
                        branch_valid.write(true);
                   } 
				   else 
				   {
          
                   sc_module::wait();
					branch_target_address.write(pc_reg + 1);
                    decode_valid.write(false);
                   sc_module::wait();
                    break;
			case 0x12:	// bgt R1, R2, label
                   src_A.write(0);
                   src_B.write(0);
                   alu_src.write(0);
				   alu_op.write(3);
                   decode_valid.write(true);
                   if (srcC_tmp > srcA_tmp)
				   {
						branch_target_address.write(pc_reg + label_tmp);
						br_instruction_address.write(instr_tmp);
                        branch_valid.write(true);
                   
				   }
				   else
				   {
				   }
	
                   sc_module::wait();
					branch_target_address.write(pc_reg + 1);
                    decode_valid.write(false);
                   sc_module::wait();
                    break;
			case 0x13:	// bge R1, R2, label
                   src_A.write(0);
                   src_B.write(0);
                   alu_src.write(0);
                   alu_op.write(3);
                   decode_valid.write(true);
                   if (srcC_tmp >= srcA_tmp)
				   {
						branch_target_address.write(pc_reg + label_tmp);
						br_instruction_address.write(instr_tmp);
                        branch_valid.write(true);
                   } 
					sc_module::wait(); ;
					branch_target_address.write(pc_reg + 1);
					decode_valid.write(false);
                   sc_module::wait();
                    break;
			case 0x14:	// blt R1, R2, label
                    src_A.write(0);
                    src_B.write(0);
                    alu_src.write(0);
                    alu_op.write(3);
                    decode_valid.write(true);
                    if (srcC_tmp < srcA_tmp) 
					{
						branch_target_address.write(pc_reg + label_tmp);
						br_instruction_address.write(instr_tmp);
                        branch_valid.write(true);
                                               
                    }
			
                   sc_module::wait();
					branch_target_address.write(pc_reg + 1);
                    decode_valid.write(false);
                   sc_module::wait();
                    break;
			case 0x15:	// ble R1, R2, label
                    src_A.write(0);
                    src_B.write(0);
                    alu_src.write(0);
                    alu_op.write(3);
                    decode_valid.write(true);
                    if (srcC_tmp <= srcA_tmp)
					{
						branch_target_address.write(pc_reg + label_tmp);
						br_instruction_address.write(instr_tmp);
                        branch_valid.write(true);
                    }
				   sc_module::wait();
					branch_target_address.write(pc_reg + 1);
                    decode_valid.write(false);
                   sc_module::wait();
                    break;
			case 0x16:	// j label
                    src_A.write(0);
                    src_B.write(0);
                    alu_src.write(0);
                    alu_op.write(3);
                    decode_valid.write(true);
                    branch_target_address.write(longlabel_tmp);
					br_instruction_address.write(instr_tmp);
                    branch_valid.write(true);
                   sc_module::wait();
					branch_target_address.write(pc_reg + 1);
                    decode_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
			case 0x17:	// jal label for procedure call
					sc_module::wait(); ;
					sc_module::wait(); ;
					break;
			case 0x18:	// jr
					sc_module::wait(); ;
					sc_module::wait(); ;
					break;
            case 0x29:      // fadd R1, R2, R3
                    src_A.write(srcA_tmp);
                    src_B.write(srcB_tmp);
	                alu_src.write(regC_tmp);
                    alu_op.write(3);
                    float_valid.write(true);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    float_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
             case 0x2a:      // fsub R1, R2, R3
                    src_A.write(srcA_tmp);
                    src_B.write(srcB_tmp);
                    alu_src.write(regC_tmp);
                    alu_op.write(4);
                    float_valid.write(true);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    float_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
             case 0x2b:      // fmul R1, R2, R3
                    src_A.write(srcA_tmp);
                    src_B.write(srcB_tmp);
                    alu_src.write(regC_tmp);
                    alu_op.write(5);
					float_valid.write(true);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    float_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
             case 0x2c:      // fdiv R1, R2, R3
                    src_A.write(srcA_tmp);
                    src_B.write(srcB_tmp);
                    alu_src.write(regC_tmp);
                    alu_op.write(6);
                    float_valid.write(true);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    float_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
             case 0x31:      // mmxadd R1, R2, R3
	              src_A.write(srcA_tmp);
		          src_B.write(srcB_tmp);
			      alu_src.write(regC_tmp);
				  alu_op.write(3);
				  mmx_valid.write(true);
				   work_load (INFINITE, ThreadDecodeID)  ;     sc_module::wait();
                  mmx_valid.write(false);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  break;
             case 0x32:      // mmxadds R1, R2, R3
                  src_A.write(srcA_tmp);
                  src_B.write(srcB_tmp);
                  alu_src.write(regC_tmp);
                  alu_op.write(4);
                  mmx_valid.write(true);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  mmx_valid.write(false);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  break;
             case 0x33:      // mmxsub R1, R2, R3
                  src_A.write(srcA_tmp);
                  src_B.write(srcB_tmp);
                  alu_src.write(regC_tmp);
                  alu_op.write(5);
                  mmx_valid.write(true);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  mmx_valid.write(false);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  break;
             case 0x34:      // mmxsubs R1, R2, R3
                  src_A.write(srcA_tmp);
                  src_B.write(srcB_tmp);
                  alu_src.write(regC_tmp);
                  alu_op.write(6);
                  mmx_valid.write(true);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  mmx_valid.write(false);
                 work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                  break;
             case 0x35:      // pmadd R1, R2, R3
				src_A.write(srcA_tmp);
				src_B.write(srcB_tmp);
                alu_src.write(regC_tmp);
                alu_op.write(7);
                mmx_valid.write(true);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                mmx_valid.write(false);
               work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                break;
             case 0x36:      // pack R1, R2, R3
			        src_A.write(srcA_tmp);
		            src_B.write(srcB_tmp);
	                alu_src.write(regC_tmp);
                    alu_op.write(8);
                    mmx_valid.write(true);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    mmx_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
             case 0x37:      // mmxck R1, R2, R3 MMX Chroma Keying
                                 
                    src_A.write(srcA_tmp);
                    src_B.write(srcB_tmp);
                    alu_src.write(regC_tmp);
                    alu_op.write(9);
                    mmx_valid.write(true);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    mmx_valid.write(false);
                   work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                    break;
             case 0xe0:      // flush register
                    src_A.write(0);
                    src_B.write(0);
                    alu_src.write(0);
                    alu_op.write(0);
					for (i = 0; i< 32; i++) {
						cpu_reg[i] = 0;
					}
                   work_load (INFINITE, ThreadDecodeID)  ; 
				sc_module::wait();
                   work_load (INFINITE, ThreadDecodeID)  ;  
				sc_module::wait();
                    break;
              case 0xf0:      // ldpid process_number
						src_A.write(0);
						src_B.write(0);
						alu_src.write(0);
						alu_op.write(0);
						pid_valid.write(true);
						decode_valid.write(false);
						float_valid.write(false);
						mmx_valid.write(false);
						pid_data.write(longlabel_tmp);
	                      work_load (INFINITE, ThreadDecodeID)  ;  
		    			sc_module::wait(); ;
						pid_valid.write(false);
                       work_load (INFINITE, ThreadDecodeID)  ;      sc_module::wait();
                        break;
			case 0xf1:	// movi R1, #value
		    		src_A.write(imm_tmp);
					src_B.write(0);
					alu_src.write(regC_tmp);
					alu_op.write(3);	
					decode_valid.write(true);
					sc_module::wait(); ;
					decode_valid.write(false);
					sc_module::wait(); ;
					break;
             case 0xff:      // QUIT
					decode_valid.write(false);
					float_valid.write(false);
					mmx_valid.write(false);
					sc_module::wait(); ;
				
					sc_stop();
					sc_module::wait(); ;
					sc_module::wait(); ;
					break;
			default :
				
					sc_module::wait(); ;
					break;
		}
		next_pc.write(true);
		sc_module::wait();
	} 
}
	else
	{
		next_pc.write(true);
		sc_module::wait();
	}


  }
  
} // end of entry function


