/*****************************************************************************

  The following code is derived, directly or indirectly, from the SystemC
  source code Copyright (c) 1996-2006 by all Contributors.
  All Rights reserved.

  The contents of this file are subject to the restrictions and limitations
  set forth in the SystemC Open Source License Version 2.4 (the "License");
  You may not use this file except in compliance with such restrictions and
  limitations. You may obtain instructions on how to receive a copy of the
  License at http://www.systemc.org/. Software distributed by Contributors
  under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
  ANY KIND, either express or implied. See the License for the specific
  language governing rights and limitations under the License.

 *****************************************************************************/

/*****************************************************************************
 
  fetch.cpp -- Instruction Fetch Unit.
 
  Original Author: Martin Wang, Synopsys, Inc.
 
 *****************************************************************************/
 
/*****************************************************************************
 
  MODIFICATION LOG - modifiers, enter your name, affiliation, date and
  changes you are making here.
 
      Name, Affiliation, Date:
  Description of Modification:
 
 *****************************************************************************/
#include "src/systemc.h"
#include "fetch.h"
#include "SystemC/os_systemc.h"
#include "directive.h"
#include "work_load.h"
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>

// These variables are used for priniting only.
// Printing is performed when the BIOS unit is not active.
extern "C" bool mainSystem_printfetchFlag ;
extern "C" bool  mainSystem_printfetchBranchFlag ;
extern "C" bool mainSystem_printfetchAlertFlag ; 
unsigned long addr_tmp=0;
unsigned long datao_tmp=0;
unsigned long datai_tmp=0;
unsigned long lock_tmp = 0;

void FetchPrint () 
{
	if (PRINT_IFU) 
	{
		if 	( mainSystem_printfetchFlag == true ) 
		{
			mainSystem_printfetchFlag = false ; 
			if ( mainSystem_printfetchBranchFlag == true ) 
			{
				System_printf("IFU ALERT: **BRANCH**\n");
				System_printf( "---------------------------------------------------------------------------\n" ) ;
				System_printf( "IFU :  mem=0x(%0h)",  datai_tmp,"\n" ) ;
				System_printf( "IFU : PC=(%0h) ",  addr_tmp, "\n" ) ;
			}
			else
			{

   				cout << "Instruction Fetch Unit: " << "PC= " << addr_tmp << " Mem=0x" <<  datai_tmp <<endl;
				System_printf( "---------------------------------------------------------------------------\n" ) ;
			
			}
			mainSystem_printfetchBranchFlag = false ; 
		
			if (mainSystem_printfetchAlertFlag==true)
			{
				mainSystem_printfetchAlertFlag = false ; 
				System_printf("ALERT: Instruction Fetch: **INTERRUPT**\n");
			}
		}	
	}
}




void fetch::entry_fetch()
{ 
   bool bios_valid_flag ; 
   bool icache_valid_flag ; 
   work_load ( INFINITE, ThreadFetchID) ;
   sc_module::wait() ;
   addr_tmp = 1;
   datao_tmp = 0xdeadbeef;

   reset.write( true );
   ram_cs.write(true);
   ram_we.write(false);
   fetch_address.write(addr_tmp);
   work_load ( INFINITE, ThreadFetchID) ;
   sc_module::wait ( memory_latency ) ; 		// For data to appear

   bios_valid_flag = bios_valid.read() ; 
   icache_valid_flag = icache_valid.read(); 
   while ( (( bios_valid_flag == false ) || ( icache_valid_flag == false )) )
   {
		if (icache_valid_flag==true) 
		   break ; 
		if ( bios_valid_flag == true ) 
		   break ; 	 
		  work_load ( INFINITE, ThreadFetchID) ;
		  wait() ;

		     bios_valid_flag = bios_valid.read() ; 
			 icache_valid_flag = icache_valid.read();

   } 
   if (stall_fetch.read() == true) 
   {
		datai_tmp = 0;
   } 
   else
   {
   		datai_tmp = ramdata.read();
   }

   mainSystem_printfetchFlag = true ; 


   instruction_valid.write(true);
   instruction.write(datai_tmp);
   program_counter.write(addr_tmp);
   ram_cs.write(false);
  work_load ( INFINITE, ThreadFetchID) ;  
  sc_module::wait() ;  ;
  
   instruction_valid.write(false);
   addr_tmp =  addr_tmp + 1;
  work_load ( INFINITE, ThreadFetchID) ;  wait();


  while (true) 
  {
	if (addr_tmp == 5)
	{
		reset.write(false);
		 sc_module::wait () ; //	wait() ; //  os_systemcWait(INFINITE ) ;
	}
	if (interrupt_in.read() == true)
	{

   		ram_cs.write(true);
		addr_tmp = int_vectno.read();
    	ram_we.write(false);
    	 sc_module::wait(memory_latency ) ;
    	datai_tmp = ramdata.read();
        mainSystem_printfetchAlertFlag = true ; 
		 

		instruction_valid.write(true);
		instruction.write(datai_tmp);
    	ram_cs.write(false);
		interrupt_ack.write(true);
		
		if (next_pc.read() == true) 
		{ 
			addr_tmp =  addr_tmp + 1;
		}  
		
		 sc_module::wait() ;
        instruction_valid.write(false);
		interrupt_ack.write(false);
		 sc_module::wait() ;

	}

	if (branch_valid.read() == true)
	{
		
		lock_tmp ++;
    	ram_cs.write(true);
		addr_tmp = branch_address.read();
    	ram_we.write(false);
    	 sc_module::wait (memory_latency) ; // os_systemc_ThreadWaitCycles(memory_latency ) ; //?
    	 sc_module::wait() ;
		bios_valid_flag = bios_valid.read() ; 
		icache_valid_flag = icache_valid.read(); 
		while ( (( bios_valid_flag == false ) || ( icache_valid_flag == false )) )
		 {
			if (icache_valid_flag== true) 
			  break ; 
			if ( bios_valid_flag == true ) 
			   break ; 	 
		  work_load ( INFINITE, ThreadFetchID) ;
		  sc_module::wait();
		   bios_valid_flag = bios_valid.read() ; 
		   icache_valid_flag = icache_valid.read();

		} 
    		
		datai_tmp = ramdata.read();

   		mainSystem_printfetchBranchFlag = true ;  
   	  

		instruction_valid.write(true);
		instruction.write(datai_tmp);
    	ram_cs.write(false);

		if (next_pc.read() == true) 
		{ 
			addr_tmp = addr_tmp + 1 ;
		}
		fetch_address.write(addr_tmp); // YS. 5/7/2010
	   work_load ( INFINITE, ThreadFetchID) ;
	   sc_module::wait();
        instruction_valid.write(false);
	   work_load ( INFINITE, ThreadFetchID) ;
	   sc_module::wait();
	} 
	else
		{
			lock_tmp = 0;
    		ram_cs.write(true);
    		fetch_address.write(addr_tmp);
    		ram_we.write(false);
    		wait(memory_latency ) ; // For data to appear
		    bios_valid_flag = bios_valid.read() ; 
			icache_valid_flag = icache_valid.read(); 
			while ( (( bios_valid_flag == false ) || ( icache_valid_flag == false )) )
			 {
				if (icache_valid_flag== true) 
						break ; 
				if ( bios_valid_flag == true ) 
				   break ; 	 
				 sc_module::wait() ; //  os_systemcWait(INFINITE ) ;
				bios_valid_flag = bios_valid.read() ; 

				icache_valid_flag = icache_valid.read();

			 }  

    	datai_tmp = ramdata.read();

   		mainSystem_printfetchFlag = true; 
   

        instruction_valid.write(true);
        instruction.write(datai_tmp);
   		program_counter.write(addr_tmp);
		branch_clear.write(false);
    	ram_cs.write(false);
		if (next_pc.read() == true) 
		{ 
			addr_tmp = addr_tmp + 1 ; 
		}
       work_load ( INFINITE, ThreadFetchID) ;
       sc_module::wait();
        instruction_valid.write(false);
        sc_module::wait() ;
        }
	if (lock_tmp == 1) {
		branch_clear.write(true);
		 sc_module::wait() ;
	}

  }
} // end of entry function



