/*****************************************************************************

  The following code is derived, directly or indirectly, from the SystemC
  source code Copyright (c) 1996-2006 by all Contributors.
  All Rights reserved.

  The contents of this file are subject to the restrictions and limitations
  set forth in the SystemC Open Source License Version 2.4 (the "License");
  You may not use this file except in compliance with such restrictions and
  limitations. You may obtain instructions on how to receive a copy of the
  License at http://www.systemc.org/. Software distributed by Contributors
  under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
  ANY KIND, either express or implied. See the License for the specific
  language governing rights and limitations under the License.

 *****************************************************************************/

/*****************************************************************************
 
  floating.cpp -- Floating Point Execution Unit.
 
  Original Author: Martin Wang, Synopsys, Inc.
 
 *****************************************************************************/
 
/*****************************************************************************
 
  MODIFICATION LOG - modifiers, enter your name, affiliation, date and
  changes you are making here.
 
      Name, Affiliation, Date:
  Description of Modification:
 
 *****************************************************************************/
 

#include "src/systemc.h"
#include "floating.h"
#include "directive.h"
#include "SystemC/os_systemc.h"
#include "work_load.h"
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>
  int			      	  float_opcode_tmp = 0;
  signed int		        dout_tmp = 0;
  unsigned int		        dest_tmp = 0;

  unsigned int                  fpua_sign_tmp;
  unsigned int                  fpua_exponent_tmp;
  unsigned int                  fpua_significand_tmp;
  unsigned int                  fpub_sign_tmp;
  unsigned int                  fpub_exponent_tmp;
  unsigned int                  fpub_significand_tmp;
  const char *                  opcode_encode="";
  unsigned int                  fpua_tmp;
  unsigned int                  fpub_tmp;


  int				exponent_diff_tmp = 0;
  int				exponent_tmp = 0;
  unsigned int			significant_result = 0;
  unsigned int			overflow_sign_tmp = 0;
  unsigned int			result_exp_tmp = 0;
  unsigned int			result_sign_tmp = 0;
  
  extern "C" bool mainSystem_printfloatFlag ;
  extern "C" bool mainSystem_printfloatResultsFlag ;


void FloatPrintResults () 
{
	if ( mainSystem_printfloatResultsFlag == true )
	{
		mainSystem_printfloatResultsFlag = false ; 
		System_printf("\t-------------------------------\n");
		System_printf("Floating Point Unit : Result   Sign=0x%x.\n",result_sign_tmp);
		System_printf("\t   : Result   Mantissa=0x%x.\n", significant_result);
		System_printf("\t   : Result   Exponent=0x%x.\n", result_exp_tmp);
		System_printf("\t   : Overflow Mantissa=0x%x.\n", overflow_sign_tmp, "\t   : " ,
				opcode_encode, "\n" );
		System_printf(" Floating Point Unit Output =0x%x.\n",dout_tmp);
	}
}



void  FloatPrintInputs ()
{
 	System_printf("\t\t\t\t\t\t\t-------------------------------\n");
	 	System_printf("\t\t\t\t\t\t\tFPU: A s=0x%x, exp=0x%x, mant=0x%x\n",fpua_sign_tmp, fpua_exponent_tmp, fpua_significand_tmp);
	 	System_printf("\t\t\t\t\t\t\t   : B s=0x%x, exp=0x%x, mant=0x%x\n",fpub_sign_tmp, fpub_exponent_tmp, fpub_significand_tmp);
		System_printf("\t\t\t\t\t\t\t   : exponent difference = %d\n",exponent_diff_tmp);
	 	System_printf("\t\t\t\t\t\t\t-------------------------------\n");
}

void floating::entry_float()
{

  //
  // main loop
  //
  // 
  // current implementation only worked if both operand have same sign bits
  // overflow is ignored.

  // initialization of output

	sc_module::wait() ; 
	work_load ( INFINITE, ThreadFloatID) ;
	sc_module::wait() ;  ;
	for (int ij=0; ij<3; ij++ ) 
	{
		work_load ( INFINITE, ThreadFloatID) ;
		sc_module::wait();
	}
	while(true)
  {
	  

    	while  (in_valid.read() == false)
		{
			work_load ( INFINITE, ThreadFloatID) ;
			sc_module::wait();
		}
		dest_tmp = dest.read();
 		float_opcode_tmp = opcode.read();

        fpua_tmp = floata.read();
        fpub_tmp = floatb.read();
 
        fpua_sign_tmp = (fpua_tmp & 0x80000000) >> 31 ;
        fpub_sign_tmp = (fpub_tmp & 0x80000000) >> 31 ;

        fpua_exponent_tmp = (fpua_tmp & 0x7f800000) >> 23 ;
        fpub_exponent_tmp = (fpub_tmp & 0x7f800000) >> 23 ;

       	fpua_significand_tmp = (fpua_tmp & 0x007fffff) ;
        fpub_significand_tmp = (fpub_tmp & 0x007fffff) ;

		exponent_diff_tmp = int(fpua_exponent_tmp) - int(fpub_exponent_tmp);


	 	mainSystem_printfloatFlag = true ;

		if (exponent_diff_tmp > 0) 
		{
			exponent_tmp = fpua_exponent_tmp;
		//System_printf("shift significant B to Right\n");
       		fpub_significand_tmp = fpub_significand_tmp >> exponent_diff_tmp ;
			fpub_exponent_tmp = fpua_exponent_tmp;
		} 
		else
		{
			exponent_tmp = fpub_exponent_tmp;
			//System_printf("shift significant A to Right\n");
       			fpua_significand_tmp = fpua_significand_tmp >> exponent_diff_tmp ;
			fpua_exponent_tmp = fpub_exponent_tmp;
		}
	work_load ( INFINITE, ThreadFloatID) ;
	sc_module::wait();
		mainSystem_printfloatFlag = true ;

		// output MUX
		switch (float_opcode_tmp) {
		case 0:         // Stall
			opcode_encode = "STALL";
			dout_tmp = dout_tmp;
			sc_module::wait();
			break;
		case 3:         // add
			opcode_encode = "FADD";
			significant_result = int(fpua_significand_tmp) + int(fpub_significand_tmp); 
		    sc_module::wait ()  ;
			break;
		case 4:         // sub
			opcode_encode = "FSUB";
			significant_result = int(fpua_significand_tmp) - int(fpub_significand_tmp); 
			work_load ( INFINITE, ThreadFloatID) ;
			sc_module::wait();
			break;
		case 5:         // mul
			opcode_encode = "FMUL";
			significant_result = int(fpua_significand_tmp) * int(fpub_significand_tmp); 
			fpub_exponent_tmp *= 2;			// exponent is doubled in value
			work_load ( INFINITE, ThreadFloatID) ;
			sc_module::wait();
			break;
		case 6:         // div
			opcode_encode = "FDIV";
			significant_result = int(fpua_significand_tmp) / int(fpub_significand_tmp); 
			work_load ( INFINITE, ThreadFloatID) ;
			sc_module::wait();
			break;
		default:
		 	System_printf("\t\t\t\t\t\t\t-------------------------------\n");
			System_printf("\t\t\t\t\t\t\tFPU: Bad Opcode %d.\n",float_opcode_tmp);
		 	System_printf("\t\t\t\t\t\t\t-------------------------------\n");
		work_load ( INFINITE, ThreadFloatID) ;
		sc_module::wait();
		
		break;
	}	

	overflow_sign_tmp = (significant_result & 0xff800000) >> 23;
	dout_tmp = (significant_result << overflow_sign_tmp) & 0x007fffff ;
	result_exp_tmp = fpub_exponent_tmp  + overflow_sign_tmp; 
	dout_tmp = dout_tmp | ((result_exp_tmp << 23) & 0x7f800000) ;
	result_sign_tmp = fpua_sign_tmp;
	dout_tmp = dout_tmp | ((result_sign_tmp << 31) & 0x80000000) ;
	System_printf("\t\t\t\t\t\t\t-------------------------------\n");
	System_printf("\t\t\t\t\t\t\tFPU: Result   Sign=0x%x.\n",result_sign_tmp);
	System_printf("\t\t\t\t\t\t\t   : Result   Mantissa=0x%x.\n",significant_result);
	System_printf("\t\t\t\t\t\t\t   : Result   Exponent=0x%x.\n",result_exp_tmp);
	System_printf("\t\t\t\t\t\t\t   : Overflow Mantissa=0x%x.\n",overflow_sign_tmp);
	cout << "\t\t\t\t\t\t\t   : " << opcode_encode;
	System_printf(" FPU Output =0x%x.\n",dout_tmp);
	cout <<"\t\t\t\t\t\t\t   : at CSIM " << sc_time_stamp() << endl;
	System_printf("\t\t\t\t\t\t\t-------------------------------\n");
	fdout.write(dout_tmp);
	fout_valid.write(true);
	fdestout.write(dest_tmp);
  sc_module::wait();
	fout_valid.write(false);
  sc_module::wait();
  }

}


