/*  The following code is derived, directly or indirectly, from the SystemC
 source code Copyright (c) 1996-2006 by all Contributors.
 All Rights reserved.

 The contents of this file are subject to  restrictions and limitations
 set forth in the SystemC Open Source License Version 2.4 (the "License");
 You may not use this file except in compliance with such restrictions andarb

 limitations. You may obtain instructions on how to receive a copy of the
 License at http://www.systemc.org/. Software distributed by Contributors
 under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF
 ANY KIND, either express or implied. See the License for the specific
 language governing rights and limitations under the License.

 *****************************************************************************/

/*****************************************************************************
 
 main -- This is a simple CPU modeling using SystemC.
 Architecure defined by Martin Wang.
 You can initialize register by modifying file
 named register, and so is bios, and dacache.
 
 Original Author: Martin Wang, Synopsys, Inc.
 Revised:         Dr. Yehdua Singer, Beyond2000 Ltd.
 
 *****************************************************************************/

/*****************************************************************************
 
 MODIFICATION LOG - modifiers, enter your name, affiliation, date and
 changes you are making here.
 
 Name, Affiliation, Date:
 Description of Modification:
 
 *****************************************************************************/

#include <climits>
#include "src/systemc.h"
#include "directive.h"

#include "bios.h"
#include "paging.h"
#include "icache.h"
#include "fetch.h"
#include "decode.h"
#include "exec.h"
#include "mmxu.h"
#include "floating.h"
#include "dcache.h"
#include "pic.h"

#include "display_results.h"
#include <xdc/runtime/Timestamp.h>

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>

#include <ti/sysbios/BIOS.h>

#include <ti/sysbios/knl/Task.h>


#include <stdio.h>

//#include <common/console/console.h>
//#include <common/servers/servers.h>

/* BIOS6 include */
#include <ti/sysbios/BIOS.h>
// ************************ ICACHE ***********************************
// ICACHE = ram_cs
// ICACHE = ram_we
// ICACHE = addr
// ICACHE = ram_datain
// ICACHE = ram_dataout
// ICACHE = ld_valid = pid_valid
// ICACHE = ld_data = pid_data
sc_signal<bool> icache_valid("ICACHE_VALID");

// ************************ BIOS ***********************************
sc_signal<bool> ram_cs("RAM_CS");
sc_signal<bool> ram_we("RAM_WE");
sc_signal<unsigned> addr("Address");
sc_signal<unsigned> ram_datain("RAM_DATAIN");
sc_signal<unsigned> ram_dataout("RAM_DATAOUT");
sc_signal<bool> bios_valid("BIOS_VALID");
const int delay_cycles = 2;

// ************************ Paging ***********************************
// Paging paging_din = ram_datain
// Paging paging_csin = ram_cs
// Paging paging_wein = ram_we
// Paging logical_address = addr 
sc_signal<unsigned> icache_din("ICACHE_DIN");
sc_signal<bool> icache_validin("ICACHE_VALIDIN");
sc_signal<bool> icache_stall("ICACHE_STALL");
sc_signal<unsigned> paging_dout("PAGING_DOUT");
sc_signal<bool> paging_csout("PAGING_CSOUT");
sc_signal<bool> paging_weout("PAGING_WEOUT");
sc_signal<unsigned> physical_address("PHYSICAL_ADDRESS");
// Paging dataout  = ram_dataout 
// Paging data_valid = icache_valid
// Paging stall_ifu = stall_fetch

// ************************ Fetch ***********************************
// IFU ramdata = ram_dataout
sc_signal<unsigned> branch_target_address("BRANCH_TARGET_ADDRESS");
sc_signal<bool> next_pc("NEXT_PC");
sc_signal<bool> branch_valid("BRANCH_VALID");
sc_signal<bool> stall_fetch("STALL_FETCH");
sc_signal<bool> pred_fetch("PRED_FETCH");
// IFU ram_valid = bios_valid
// IFU ram_cs = ram_cs
// IFU ram_we = ram_we
// IFU address = addr
// IFU smc_instrction = ram_datain
// IFU pred_branch_address = pred_branch_address
// IFU pred_branch_valid = pred_branch_valid
sc_signal<unsigned> instruction("INSTRUCTION");
sc_signal<bool> instruction_valid("INSTRUCTION_VALID");
sc_signal<unsigned> program_counter("PROGRAM_COUNTER");
sc_signal<bool> branch_clear("BRANCH_CLEAR");
sc_signal<bool> pred_fetch_valid("PRED_FETCH_VALID");
sc_signal<bool> reset("RESET");

// ************************ Branch ***********************************
// BPU: fetch_inst = instruction
// BPU: fetch_pc = program_counter
// BPU: fetch_valid = instruction_valid
// BPU: branch_inst_addr = branch_instruction_address
// BPU: branch_target_address = branch_target_address
// BPU: branch_valid = branch_valid
sc_signal<unsigned> pred_branch_address("PRED_BRANCH_ADDRESS");
sc_signal<bool> pred_branch_valid("PRED_BRANCH_VALID");
sc_signal<bool> pred_tellid("PRED_TELLID");
sc_signal<unsigned> pred_instruction("PRED_INSTRUCTION");
sc_signal<bool> pred_inst_valid("PRED_INST_VALID");
sc_signal<unsigned> pred_inst_pc("PRED_INST_PC");

// ************************ Decode ***********************************
// ID instruction = instruction
// ID instruction = instruction_valid
// ID destreg_write = out_valid
// ID destreg_write_src = destout
// ID clear_branch     = branch_clear 
// ID pc = program_counter
sc_signal<bool> pred_on("PRED_ON");
sc_signal<unsigned> branch_instruction_address("BR_INSTRUCTION_ADDRESS");
// ID alu_dataout = dout from EXEC 
sc_signal<signed> dram_dataout("DRAM_DATAOUT");
sc_signal<bool> dram_rd_valid("DRAM_RD_VALID");
sc_signal<unsigned> dram_write_src("DRAM_WRITE_SRC");
// ID next_pc     = next_pc
// ID branch_valid = branch_valid
// ID branch_target_address = branch_target_address
sc_signal<bool> mem_access("MEM_ACCESS");
sc_signal<unsigned> mem_address("MEM_ADDRESS");
sc_signal<int> alu_op("ALU_OP");
sc_signal<bool> mem_write("MEM_WRITE");
sc_signal<unsigned> alu_src("ALU_SRC");
sc_signal<bool> reg_write("REG_WRITE");
sc_signal<signed int> src_A("SRC_A");
sc_signal<signed int> src_B("SRC_B");
sc_signal<bool> forward_A("FORWARD_A");
sc_signal<bool> forward_B("FORWARD_B");
// ID stall_fetch = stall_fetch
sc_signal<bool> decode_valid("DECODE_VALID");
sc_signal<bool> float_valid("FLOAT_VALID");
sc_signal<bool> mmx_valid("MMX_VALID");
sc_signal<bool> pid_valid("PID_VALID");
sc_signal<signed> pid_data("PID_DATA");

// ************************ DCACHE  ***********************************
sc_signal<signed> mmic_datain("MMIC_DATAIN"); /* DCU: datain 	*/
sc_signal<unsigned> mmic_statein("MMIC_STATEIN");/* DCU: statein */
sc_signal<bool> mmic_cs("MMIC_CS"); /* DCU: cs 	*/
sc_signal<bool> mmic_we("MMIC_WE"); /* DCU: we 	*/
sc_signal<unsigned> mmic_addr("MMIC_ADDR"); /* DCU: addr	*/
sc_signal<unsigned> mmic_dest("MMIC_DEST"); /* DCU: dest 	*/
sc_signal<unsigned> mmic_destout("MMIC_DESTOUT");/* DCU: destout */
sc_signal<signed> mmic_dataout("MMIC_DATAOUT");/* DCU: dataout */
sc_signal<bool> mmic_out_valid("MMIC_OUT_VALID");/* DCU: out_valid*/
sc_signal<unsigned> mmic_stateout("MMIC_STATEOUT");/* DCU: stateout */

// ************************ Execute ***********************************
// EXEC in_valid = decode_valid
sc_signal<bool> in_valid("IN_VALID");
// EXEC opcode = alu_op
sc_signal<bool> negate("NEGATE");
sc_signal<int> add1("ADD1");
sc_signal<bool> shift_sel("SHIFT_SEL");
// EXEC dina = src_A
// EXEC dinb = src_B
// EXEC dest = alu_src
sc_signal<bool> c("C");
sc_signal<bool> v("V");
sc_signal<bool> z("Z");
sc_signal<signed> dout("DOUT");
sc_signal<bool> out_valid("OUTPUT_VALID");
sc_signal<unsigned> destout("DESTOUT");

// ************************ Floating point ******************************
// FPU in_valid = float_valid
// FPU opcode = alu_op
// FPU floata = src_A
// FPU floatb = src_B
// FPU dest = alu_src
sc_signal<signed> fdout("FDOUT");
sc_signal<bool> fout_valid("FOUT_VALID");
sc_signal<unsigned> fdestout("FDESTOUT");

// ************************ PIC *****************************************
sc_signal<bool> ireq0("IREQ0");
sc_signal<bool> ireq1("IREQ1");
sc_signal<bool> ireq2("IREQ2");
sc_signal<bool> ireq3("IREQ3");
// PIC cs = interrupt_ack
// PIC intack_cpu = interrupt_ack
sc_signal<bool> rd_wr("RD_WR");
sc_signal<bool> intreq("INTREQ");
sc_signal<unsigned> vectno("VECTNO");
sc_signal<bool> intack("INTACK");
sc_signal<bool> intack_cpu("INTACK_CPU");

// ************************ MMX ***********************************
// MMX mmx_valid = mmx_valid
// MMX opcode = alu_op
// MMX mmxa = src_A
// MMX mmxb = src_B
// MMX dest = dest
// MMX mmxdout = fdout
// MMX mmxout_valid = fpu_valid
// MMX mmxdestout = fpu_destout 

// ************************ DSP *****************************************
sc_signal<int> dsp_in1("DPS_IN1");
sc_signal<int> dsp_out1("DSP_OUT1");
sc_signal<bool> dsp_data_valid("DSP_DATA_VALID");
sc_signal<bool> dsp_input_valid("DSP_INPUT_VALID");
sc_signal<bool> dsp_data_requested("DSP_DATA_REQUESTED");

//  const int delay_cycles = 2;
#define MAX_CODE_LENGTH 500
///

long lSchedCount;

UInt32 ProcessTime_start;
UInt32 ProcessTime_end;

int ProcessTimeHours;
int ProcessTimeMinutes;
int ProcessTimeSeconds;
int ProcessTimeMilliSeconds;

void DachePrintRead();
void DcachePrintWrite();
void DecodePrintDumpedRegist();
void FloatPrintResults();
void IcachePrintAlertReset();
void IcachePrintIcachePid();

void main_GetSystemTimeStart() {
	ProcessTime_start = Timestamp_get32();
	ProcessTime_end = ProcessTime_start;

}

void main_GetSystemTimeEnd() {

	ProcessTime_end = Timestamp_get32();
}
void main_InitTime() {
	ProcessTimeHours = 0;
	ProcessTimeMinutes = 0;
	ProcessTimeSeconds = 0;
	ProcessTimeMilliSeconds = 0;
	ProcessTime_start = ProcessTime_end;

}

void main_caclulate_process_time() {
	int timeHours;
	int timeMinutes;
	int timeSeconds;
	int timeMilliSeconds;
	float timeCycles;
	unsigned int deltaTimeStamp;
	deltaTimeStamp = ProcessTime_end - ProcessTime_start;
	timeCycles = deltaTimeStamp / 5.0e8;
	timeHours = timeCycles / 3600;
	timeCycles = timeCycles - timeHours * 3600;
	timeMinutes = timeCycles / 60;
	timeCycles = timeCycles - timeMinutes * 60;
	timeSeconds = timeCycles;
	timeCycles = timeCycles - timeSeconds * 60;
	timeMilliSeconds = timeCycles;

	if (timeMilliSeconds < 0) {
		timeMilliSeconds = timeMilliSeconds + 1000;
		timeSeconds = timeSeconds - 1;
	}
	if (timeSeconds < 0) {
		timeSeconds = timeSeconds + 60;
		timeMinutes = timeMinutes - 1;

	}

	if (timeMinutes < 0) {
		timeMinutes = timeMinutes + 60;
		timeHours = timeHours - 1;

	}

	ProcessTimeHours = ProcessTimeHours + timeHours;

	ProcessTimeMinutes = ProcessTimeMinutes + timeMinutes;

	ProcessTimeSeconds = ProcessTimeSeconds + timeSeconds;

	ProcessTimeMilliSeconds = ProcessTimeMilliSeconds + timeMilliSeconds;

	if (ProcessTimeMilliSeconds >= 1000) {
		ProcessTimeSeconds = ProcessTimeSeconds + 1;
		ProcessTimeMilliSeconds = ProcessTimeMilliSeconds - 1000;

	}
	if (ProcessTimeSeconds >= 60) {
		ProcessTimeMinutes = ProcessTimeMinutes + 1;
		ProcessTimeSeconds = ProcessTimeSeconds - 60;

	}

	if (ProcessTimeMinutes >= 60) {
		ProcessTimeHours = ProcessTimeHours + 1;
		ProcessTimeMinutes = ProcessTimeMinutes - 60;
	}
}

void mainPrintPerformance() {

	main_caclulate_process_time();
	System_printf("CPU Time: ");

	if (ProcessTimeHours > 0) {
		System_printf("Hours: (%0d)" , ProcessTimeHours) ;
	}

	System_printf(" Minutes: (%0d)", ProcessTimeMinutes ) ;
	System_printf(" Seconds: (%0d)", ProcessTimeSeconds ) ;

	System_printf(" Milli - Seconds: (%0d)" , ProcessTimeMilliSeconds, "\n" ) ;
}
void mainPrintTimeStamp() {


	System_printf("\n SystemC Time(%0d) " ,  sc_time_stamp() ,  "\n" ) ;
	System_printf( "-----------------------\n");
}

unsigned long dwThreadPrintManager;
Task_Handle hThreadPrintManager;
Semaphore_Handle hThreadEventPrintManager;

extern "C" bool mainPrintBiosFlag;
extern "C" bool mainSystem_printfetchFlag;
extern "C" bool mainSystem_printfetchBranchFlag;
extern "C" bool mainSystem_printfetchAlertFlag;
extern "C" bool mainPrintDecodeFlag;
extern "C" bool mainPrintDecodeShutDownFlag;
extern "C" bool mainPintDecodeOutValidFlag;
extern "C" bool mainPrintDecodeDumpRegistersFlag;
extern "C" bool mainDecodePintOutValidFlag;
extern "C" bool mainDcacheWriteSystem_printflag;
extern "C" bool mainDcacheReadSystem_printflag;
extern "C" bool mainExecPrintDivideAlertFlag;
extern "C" bool mainSystem_printfloatFlag;
extern "C" bool mainSystem_printfloatResultsFlag;

extern "C" bool mainPrintIcacheFlag;
extern "C" bool mainPrintIcacheAlertResetFlag;
extern "C" bool mainPrintIcachePidFlag;
extern "C" bool mainIcacheOutOfRangeFlag;
extern "C" bool mainPrintPagingFlag;
extern "C" bool mainPrintExecBadCodeFlag;
extern "C" bool mainExecSystem_printflag;
extern "C" bool mainPrintPerformanceFlag;

bool mainPrintBiosFlag = false;
bool mainSystem_printfetchFlag = false;
bool mainSystem_printfetchBranchFlag = false;
bool mainSystem_printfetchAlertFlag;
bool mainPrintDecodeFlag = false;

bool mainPrintDecodeDumpRegistersFlag = false;
bool mainDecodePintOutValidFlag = false;
bool mainDcacheReadSystem_printflag = false;
bool mainDcacheWriteSystem_printflag = false;
bool mainExecPrintDivideAlertFlag = false;
bool mainSystem_printfloatFlag = false;

bool mainSystem_printfloatResultsFlag = false;
bool mainPrintIcacheFlag = false;
bool mainPrintIcacheAlertResetFlag = false;
bool mainPrintIcachePidFlag = false;
bool mainIcacheOutOfRangeFlag = false;
bool mainPrintPagingFlag = false;
bool mainPrintExecBadCodeFlag = false;
bool mainExecSystem_printflag = false;

bool mainPrintPerformanceFlag = false;

void main_SetEvent_print() {
	Semaphore_post(hThreadEventPrintManager);

}

bool mainPrintDecodeShutDownFlag = false;
bool decode_mem_access_tmp;
bool decode_mem_write_tmp;
bool decode_next_pc_tmp;
bool decode_valid_tmp;
signed int decode_srcA_tmp = 0;
signed int decode_srcB_tmp = 0;
int decode_alu_op;
signed int decode_alu_src;
unsigned int decode_mem_address_tmp;
bool decode_branch_valid_tmp;
bool decode_float_valid_tmp;
bool decode_mmx_valid_tmp;
unsigned int decode_branch_target_address_tmp;
unsigned int decode_branch_instruction_address_tmp;
unsigned int decode_pid_data_tmp;
bool decode_pid_valid_tmp;

unsigned int dcache_mmic_dataout_tmp;
unsigned int dcache_mmic_stateout_tmp;
unsigned int dcache_mmic_destout_tmp;
bool dcache_mmic_out_valid_tmp;
bool dcache_out_valid_tmp;

unsigned int exec_dout_tmp = 0;
bool exec_out_valid_tmp;
unsigned int exec_dest_tmp = 0;
bool exec_v_tmp;
bool exec_z_tmp;

signed int float_fdout_tmp;
bool float_fout_valid_tmp;
signed int float_fdestout_tmp;
bool icache_valid_tmp;
unsigned int icache_din_tmp;
unsigned int paging_dout_tmp;
bool paging_csout_tmp;
bool paging_weout_tmp;
bool paging_icache_valid_tmp;
unsigned int paging_physical_address_tmp;

signed int mmx_dout;
bool mmx_fout_valid_tmp;
unsigned int mmx_fdestout;

bool pic_intreq_tmp;
unsigned pic_vectno_tmp;

bool ireq0_tmp = false;
bool ireq1_tmp = false;
bool ireq2_tmp = false;
bool ireq3_tmp = false;

unsigned int Decode_destout;

void FetchPrint();
void biosPrint();
void decode_print_variables();
void DecodePrintDumpedRegisters();
void DcachePrintRead();
void DcachePrintWrite();
void exec_print_variables();
void FloatPrintInputs();
void IcachePrintOutOfRange();
void PagingPrint();

void mainprintExec() {
	cout.setf(ios::hex, ios::basefield);
	if (mainExecSystem_printflag == true) {
		mainExecSystem_printflag = false;

		if (mainExecPrintDivideAlertFlag == true) {
			mainExecPrintDivideAlertFlag = false;
			System_printf("Division Exception - Divide by zero \n");
		} else {
			if (mainPrintExecBadCodeFlag == true) {
				mainPrintExecBadCodeFlag = false;
				System_printf("ALU:      Bad Opcode %0x.\n");

			} else {

				if (PRINT_IEU) {

					exec_print_variables();

				}
			}
		}
		cout << endl;
	}

}

void mainSystem_printfloat() {
	if (mainSystem_printfloatFlag == true) {
		mainSystem_printfloatFlag = false;

		if (PRINT_FPU) {
			FloatPrintInputs();
		}

		if (mainSystem_printfloatResultsFlag == true) {
			FloatPrintResults();
		}
	}
}

void mainPrintIcache() {
	if (mainPrintIcacheFlag == true) {
		mainPrintIcacheFlag = false;

		if (PRINT_ICU) {
			System_printf("Instruction Cache Unit: Fetching mem[%0x] ",
					physical_address.read());
			if (addr.read() < IcacheCodeSize)
				System_printf("Memory (%0x)", paging_dout.read());

			cout << endl;

		} else {
			System_printf("ICU ALERT: **MEMORY OUT OF RANGE**\n");
		}

	}
}

void mainPrintDecodeShutDown() {
	if ((mainPrintDecodeShutDownFlag == true) & (PRINT_ID == true)) {
		mainPrintDecodeShutDownFlag = false;

		System_printf("\tID: - SHUTDOWN - ");
		System_printf("\tID: - PLEASE WAIT ...... - \n");
		cout << "\t-------------------------------" << endl;
		mainPrintDecodeDumpRegistersFlag = true;
		DecodePrintDumpedRegisters();
		mainPrintPerformanceFlag = true;

	}
}

void mainPrintManager() {
	bool System_printflag;

	while (true) {

		Semaphore_pend(hThreadEventPrintManager, BIOS_WAIT_FOREVER);

		cout.setf(ios::hex, ios::basefield);
		System_printflag = mainSystem_printfetchFlag || mainPrintDecodeDumpRegistersFlag
				|| mainPrintDecodeFlag || mainSystem_printfetchAlertFlag
				|| mainPrintBiosFlag || mainPrintPagingFlag || mainExecSystem_printflag
				|| mainSystem_printfloatFlag || mainExecPrintDivideAlertFlag
				|| mainPrintExecBadCodeFlag || mainSystem_printfloatResultsFlag
				|| mainPrintIcacheAlertResetFlag || mainPrintIcachePidFlag
				|| mainPrintIcacheFlag || mainDcacheWriteSystem_printflag
				|| mainDcacheReadSystem_printflag || mainIcacheOutOfRangeFlag
				|| mainPrintDecodeShutDownFlag;

		if (System_printflag == true) {

			if (PRINT_TIME_STAMP == true)
				mainPrintTimeStamp();
			biosPrint();
			FetchPrint();
			decode_print_variables();
			mainprintExec();

			PagingPrint();
			mainprintExec();
			mainSystem_printfloat();
			IcachePrintAlertReset();
			IcachePrintIcachePid();
			mainPrintIcache();
			IcachePrintOutOfRange();
			DachePrintRead();
			DcachePrintWrite();
			mainPrintDecodeShutDown();
		}

		if (mainPrintPerformanceFlag == true) {
			mainPrintPerformanceFlag = false;
			mainPrintPerformance();
		}
	}

}

void mainPrintmanagerInit() {

	hThreadPrintManager = Task_create((Task_FuncPtr) mainPrintManager, NULL,
			NULL); // &dwThreadPrintManager) ; 

	hThreadEventPrintManager = Semaphore_create(0, NULL, NULL);

}

int sc_main(int ac, char *av[]) {

	////////////////////////////////////////////////////////////////////////////
	// 				MAIN PROGRAM 
	////////////////////////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////////////
	// 				MAIN PROGRAM 
	////////////////////////////////////////////////////////////////////////////

	sc_clock clk("Clock", 1, SC_NS, 0.5, 0.0, SC_NS);

	System_printf(
			"/////////////////////////////////////////////////////////////////////////\n");
	System_printf("//  This code is written at SYNOPSYS, Inc.\n");
	System_printf(
			"/////////////////////////////////////////////////////////////////////////\n");
	System_printf("//  Module   : main of CPU Model\n");
	System_printf("//  Author   : Martin Wang\n");
	System_printf("//  Company  : SYNOPSYS, Inc.\n");
	System_printf("//  Purpose  : This is a simple CPU modeling using SystemC.\n");
	System_printf(
			"//             Instruction Set Architecure defined by Martin Wang.\n");
	System_printf("//             \n");
	System_printf(
			"//           SystemC (TM) Copyright (c) 1988-2001 by Synopsys, Inc.  \n");
	System_printf("//             \n");
	System_printf("// Revised program to Multi-Task Environment \n");
	System_printf("//                  Dr. Yehuda Singer  \n");
	System_printf(			"/////////////////////////////////////////////////////////////////////////\n");
	System_printf("//  IN THIS MACHINE Integer is " , sizeof(int),  " bytes.\n") ;
	System_printf( "// IN THIS MACHINE Floating is " ,  sizeof(float),  " bytes.\n");
	System_printf("//  IN THIS MACHINE Double is " , sizeof(double), " bytes.\n") ;
	System_printf("//     \n");
	System_printf("//     \n");
	System_printf("//                            .,,uod8B8bou,,.\n");
	System_printf("//                   ..,uod8BBBBBBBBBBBBBBBBRPFT?l!i:.\n");
	System_printf("//              ,=m8BBBBBBBBBBBBBBBRPFT?!||||||||||||||\n");
	System_printf("//              !...:!TVBBBRPFT||||||||||!!^^\"\"    ||||\n");
	System_printf("//              !.......:!?|||||!!^^\"\"'            ||||\n");
	System_printf("//              !.........||||        ###  #  #    ||||\n");
	System_printf("//              !.........||||  ###  #  #  #  #    ||||\n");
	System_printf("//              !.........|||| #     #  #  #  #    ||||\n");
	System_printf("//              !.........|||| #     # #   #  #    ||||\n");
	System_printf("//              !.........|||| #     ##    #  #    ||||\n");
	System_printf("//              !.........|||| #     #     ###     ||||\n");
	System_printf("//              `.........|||| #   # #            ,||||\n");
	System_printf("//               .;.......||||  ###          _.-!!|||||\n");
	System_printf("//        .,uodWBBBBb.....||||       _.-!!|||||||||!:'\n");
	System_printf("//     !YBBBBBBBBBBBBBBb..!|||:..-!!|||||||!iof68BBBBBb....\n");
	System_printf("//     !..YBBBBBBBBBBBBBBb!!||||||||!iof68BBBBBBRPFT?!::   `.\n");
	System_printf("//     !....YBBBBBBBBBBBBBBbaaitf68BBBBBBRPFT?!:::::::::     `.\n");
	System_printf(
			"//     !......YBBBBBBBBBBBBBBBBBBBRPFT?!::::::;:!^\"`;:::       `.\n");
	System_printf(
			"//     !........YBBBBBBBBBBRPFT?!::::::::::^''...::::::;         iBBbo.\n");
	System_printf(
			"//     `..........YBRPFT?!::::::::::::::::::::::::;iof68bo.      WBBBBbo.\n");
	System_printf(
			"//       `..........:::::::::::::::::::::::;iof688888888888b.     `YBBBP^'\n");
	System_printf(
			"//         `........::88::::::::::::;iof688888888888888888888b.     `\n");
	System_printf(
			"//           `......::81:::::;iof688888888888888888888888888888b.\n");
	System_printf(
			"//             `....:::;iof688888888888888888888888888888888899fT!\n");
	System_printf(
			"//               `..::!8888888888888888888888888888888899fT|!^\"'\n");
	System_printf("//                 `' !!988888888888888888888888899fT|!^\"'\n");
	System_printf("//                     `!!8888888888888888899fT|!^\"'\n");
	System_printf("//                       `!988888888899fT|!^\"'\n");
	System_printf("//                         `!9899fT|!^\"'\n");
	System_printf("//                           `!^\"'\n");
	System_printf("//     \n");
	System_printf("//     \n");
	System_printf(
			"/////////////////////////////////////////////////////////////////////////\n\n\n");

	fetch IFU("FETCH_BLOCK");
	IFU.init_param(delay_cycles);
	IFU << ram_dataout << branch_target_address << next_pc << branch_valid
			<< stall_fetch << intreq << vectno << bios_valid << icache_valid
			<< pred_fetch << pred_branch_address << pred_branch_valid << ram_cs
			<< ram_we << addr << ram_datain << instruction << instruction_valid
			<< program_counter << intack_cpu << branch_clear << pred_fetch_valid
			<< reset << clk;

	decode IDU("DECODE_BLOCK"); // IDU.decode_init() ;
	IDU << reset << instruction << pred_instruction << instruction_valid
			<< pred_inst_valid << out_valid << destout << dout << dram_dataout
			<< dram_rd_valid << destout << fdout << fout_valid << fdestout
			<< branch_clear << dsp_data_valid << program_counter << pred_on
			<< branch_instruction_address << next_pc << branch_valid
			<< branch_target_address << mem_access << mem_address << alu_op
			<< mem_write << alu_src << reg_write << src_A << src_B << forward_A
			<< forward_B << stall_fetch << decode_valid << float_valid
			<< mmx_valid << pid_valid << pid_data << clk;

	exec IEU("EXEC_BLOCK");
	IEU << reset << decode_valid << alu_op << negate << add1 << shift_sel
			<< src_A << src_B << forward_A << forward_B << alu_src << c << v
			<< z << dout << out_valid << destout << clk;

	floating FPU("FLOAT_BLOCK"); // order dependent
	FPU << float_valid << alu_op << src_A << src_B << alu_src << fdout
			<< fout_valid << fdestout << clk;
	mmxu MMXU("MMX_BLOCK");
	MMXU << mmx_valid << alu_op << src_A << src_B << alu_src << fdout
			<< fout_valid << fdestout << clk;
	bios BIOS("BIOS_BLOCK");
	BIOS.init_param(delay_cycles);
	BIOS.datain(ram_datain); // order independent
	BIOS.cs(ram_cs);
	BIOS.we(ram_we);
	BIOS.addr(addr);
	BIOS.dataout(ram_dataout);
	BIOS.bios_valid(bios_valid);
	BIOS.stall_fetch(stall_fetch);
	BIOS.CLK(clk);

	paging PAGING("PAGING_BLOCK");
	PAGING << ram_datain << ram_cs << ram_we << addr << icache_din
			<< icache_validin << icache_stall << paging_dout << paging_csout
			<< paging_weout << physical_address << ram_dataout << icache_valid
			<< stall_fetch << clk;

	icache ICACHE("ICACHE_BLOCK");
	ICACHE.init_param(delay_cycles);
	ICACHE << paging_dout << paging_csout << paging_weout << physical_address
			<< pid_valid << pid_data << icache_din << icache_validin
			<< icache_stall << clk;

	dcache DCACHE("DCACHE_BLOCK");
	DCACHE.init(delay_cycles);
	DCACHE << mmic_datain << mmic_statein << mmic_cs << mmic_we << mmic_addr
			<< mmic_dest << mmic_destout << mmic_dataout << mmic_out_valid
			<< mmic_stateout << clk;

	pic APIC("PIC_BLOCK");
	APIC << ireq0 << ireq1 << ireq2 << ireq3 << intack_cpu << rd_wr
			<< intack_cpu << intreq << intack << vectno;
	display_results DR("Display_Results_Block");
	DR.CLK(clk);
	BIOS_start();

	sc_start();

	if (mainPrintDecodeShutDownFlag == false)
		Task_sleep(12000);
	else
		Task_sleep(14000);

	return 0; /* this is necessary */
}
