#include "src/systemc.h"
#include "sysc/kernel/sc_event.h"
#include "sysc/kernel/sc_kernel_ids.h"
#include "sysc/kernel/sc_macros_int.h"
#include "sysc/kernel/sc_module.h"
#include "sysc/kernel/sc_module_registry.h"
#include "sysc/kernel/sc_name_gen.h"
#include "sysc/kernel/sc_object_manager.h"
#include "sysc/kernel/sc_cthread_process.h"
#include "sysc/kernel/sc_method_process.h"
#include "sysc/kernel/sc_thread_process.h"
#include "sysc/kernel/sc_process_handle.h"
#include "sysc/kernel/sc_simcontext.h"
#include "sysc/kernel/sc_simcontext_int.h"
#include "sysc/kernel/sc_reset.h"
#include "sysc/kernel/sc_ver.h"
#include "SystemC/os_systemc.h"

Semaphore_Handle os_systemc_hThreadSemaphoresArray[ThreadNum];
 UInt os_systemc_hThreadEventsArray[ThreadNum];

Task_Handle os_systemc_hThreadArray[ThreadNum];
unsigned long os_systemc_dwThreadIDArray[ThreadNum];

sc_core::sc_thread_handle os_systemc_HandleCThreadArray[ThreadNum];
sc_core::sc_thread_handle os_systemc_HandleThreadArray[ThreadNum];

sc_core::sc_method_handle os_systemc_HandleMThreadArray[ThreadNum];
sc_core::SC_ENTRY_FUNC os_systemc_semantics_method_p[ThreadNum];

sc_core::sc_simcontext* os_systemc_sc_simcontext_ptr[ThreadNum];
sc_core::sc_port_registry* os_systemc_sc_port_registry_ptr[ThreadNum];

sc_core::sc_module* os_systemc_module_registry_ptr[ThreadNum];
bool os_systemc_ThreadInitFlag[ThreadNum];

// Internal variables for DEBUG only. 

long os_systemc_m_lThreadWaitCount[ThreadNum];
long os_systemc_m_lThreadWorkCount[ThreadNum];
long os_systemc_wait_cycles[ThreadNum];
Event_Handle os_systemc_Event;

int os_systemc_int_ActualThreadNum = 0;

void os_systemcWait(long duration) {
	int id;
	id = os_systemc_GetCurrentThreadIndex();
	// Mark the thread as wait for SystemC.

	if (os_systemc_HandleCThreadArray[id] != NULL) {

		os_systemc_HandleCThreadArray[id]->m_fThreadActive = false;
		os_systemc_HandleCThreadArray[id]->m_lThreadWaitCount =
				os_systemc_HandleCThreadArray[id]->m_lThreadWaitCount + 1;
		os_systemc_m_lThreadWaitCount[id] =
				os_systemc_HandleCThreadArray[id]->m_lThreadWaitCount;
		os_systemc_m_lThreadWorkCount[id] =
				os_systemc_HandleCThreadArray[id]->m_lThreadWorkCount;

		os_systemc_HandleCThreadArray[id]->wait_cycles(duration);
	} else {
		if (os_systemc_HandleThreadArray[id] != NULL) {
			os_systemc_HandleThreadArray[id]->m_lThreadWaitCount =
					os_systemc_HandleThreadArray[id]->m_lThreadWaitCount + 1;
			os_systemc_HandleThreadArray[id]->m_fThreadActive = false;
			os_systemc_m_lThreadWaitCount[id] =
					os_systemc_HandleThreadArray[id]->m_lThreadWaitCount;
			os_systemc_m_lThreadWorkCount[id] =
					os_systemc_HandleThreadArray[id]->m_lThreadWorkCount;
		}

	}

	// !Add reset event to make the event non signaled (in waitable state)

	Event_post(os_systemc_Event, os_systemc_hThreadEventsArray[id]);
	Semaphore_pend(os_systemc_hThreadSemaphoresArray[id], BIOS_WAIT_FOREVER);

} // End of os_systemcWait.

void os_systemc_ThreadWaitCycles(long thread_wait_cycles) {
	int id;
	id = os_systemc_GetCurrentThreadIndex();
	os_systemc_wait_cycles[id] = thread_wait_cycles;

	// while ( os_systemc_wait_cycles[id] > 0 ) 
	// {
	os_systemcWait(thread_wait_cycles);
	//	 os_systemc_wait_cycles[id] = os_systemc_wait_cycles[id] - 1 ; 
	// }

}

int os_systemc_GetCurrentThreadIndex() {
	int i;

	for (i = 0; i < os_systemc_int_ActualThreadNum; i++) {
		//		if ( os_systemc_module_registry_ptr[i] == 3 )
		return i;
	}
	return (-i);
}

int os_systemc_map_method_or_cthread_to_osthread(
		sc_core::SC_ENTRY_FUNC method_p) {

	int i;
	if (os_systemc_int_ActualThreadNum == 0) {
		return (0);
	} else {
		for (i = 0; i < os_systemc_int_ActualThreadNum; i++) {
			if (os_systemc_semantics_method_p[i] == method_p) {

				return i;
			}

		}
	}
	return (-i);
}

void os_systemc_wait_for_all_threads_termination() {

	UInt Event_pend_result;
	do {
		Event_pend_result = Event_pend(os_systemc_Event, Event_Id_NONE, 0xffff,
				BIOS_WAIT_FOREVER);
	} while (Event_pend_result != 0xffff);

}
void os_systemc_setSemaphore(Semaphore_Handle Semaphore) {
	Semaphore_post(Semaphore);

}

bool os_systemc_ThreadCheckInit() {
	int i;
	bool ThreadInitFlag = true;

	for (i = 0; i < os_systemc_int_ActualThreadNum; i++) {
		ThreadInitFlag = ThreadInitFlag & os_systemc_ThreadInitFlag[i];
	}

	return (ThreadInitFlag);

}

void os_systemcWaitThreadSync() {

	while (os_systemc_ThreadCheckInit() == false)
		Task_sleep(10);

	// Wait for the first event.
	os_systemcWait(INFINITE);

}
